#!/usr/bin/perl -w
# Copyright Mark Zealey 2002. Released under the GPL.
use strict;
use MSN;
use IO::Handle;
use Curses;
use Text::Wrap;
use POSIX qw/strftime/;

# Most of this should be in the config file...
my $TIME_FORMAT = '';			# The format of the timestamp you want
my $MAX_FUSER_LEN = 30;			# Maximum length of a fusername.
my $PING_SAMPLE = 60;			# Send a ping every X seconds
my $PING_COUNT = 4;			# Start doing the counting thing if no
					# ping response after X seconds.
my $MAX_BUFFER_LEN = 1000;		# Length scrollback buffer (per-window)
my $MAX_LINE_LEN = 50;			# Max length of the input history
my $BEEP = 1;				# Beep on error etc ?
my $RESIZE_PROG = '/usr/X11R6/bin/resize';	# Program to get new window size


my $CONFIG = 'config';
my $MSN_VERSION = '0.5.6';

my $time_hires;
BEGIN {
	eval 'use Time::HiRes qw/time/';
	$time_hires = 1 unless $@
}

my $startup_time = time;		# Time we started
my $session = -1;			# Number of the current session
my %sessions;				# Info about all sessions
my ($prompt, $ask);
my $last_msg = '';
my $status;				# What's displayed in the status-bar
my $escape = 0;				# Status of the escape key
my $disconnect = 0;			# Disconnect when this is 1
my $hide_input = 0;			# Hide input from being displayed?
my $last_type = time;			# Contains the last time the user typed
my $last_ping = '??';			# Time taken for the last ping
my $ping_last;				# When we sent our last ping
my $ping_got = 0;			# When we got our last ping
my @lines;				# List of the lines we've read from the
					# input
my $line_ptr = 0;			# Pointer to the elem in @lines we are
					# ref'ing
my $char_ptr = 0;			# Current pos we are on the line
my $replace_on = 0;			# Do we replace or insert stuff?
my $needs_resize = 0;			# Need to resize the windows etc?
my $idle = 0;				# We idle?
my $last_user = 0;			# The last time we updated the user box

my (%conf, $win, $user_win, $in_line, $stat_line, $MAGENTA, $RED, $BLUE);

# This is the big array, defining all the /command's we can handle. It's fairly
# simple, in each array element, the following info is contained:
#
# element | description
# --------+---------------------------------------------------------------------
#       1 | An array-ref to an array of names this function is called. The first
#         | element in this sub-array is the name of the function (without the
#         | handle_ bit on the front) that will be called when any of the names
#         | in the list is typed
# --------+---------------------------------------------------------------------
#       2 | An array-ref of options. Each option must be preceeded by its
#         | required-ness, ie, a string containing some of:
#         |             REQ (required)
#         |             OPT (optional)
#         |             MUL (multiple)
# --------+---------------------------------------------------------------------
#       3 | A string containing the help text for the functions
my @funcs = (
		[
			[qw/away/],
			[],
			'Change your status to "away"'
		], [
			[qw/hide hidden/],
			[],
			'Change your status to "hidden"'
		], [
			[qw/status/],
			[
				'REQ' => 'status'
			],
			'Change your status to "[1]"'
		], [
			[qw/online/],
			[],
			'Change your status to "online"'
		], [
			[qw/window session/],
			[
				'REQ' => 'session'
			],
			'Switches to session [1]'
		], [
			[qw/msg/],
			[
				'REQ' => 'session',
				'REQ MUL' => 'text'
			],
			'Send [2] to session [1]'
		], [
			[qw/quit exit/],
			[],
			'Quits MSN'
		], [
			[qw/part bye/],
			[
				'OPT' => 'session'
			],
			'Leave [1], or the current one if none is specified'
		], [
			[qw/add/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to add [1] to your buddies list, so you '.
			'will be informed when they are online etc...'
		], [
			[qw/del/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to remove [1] from your buddies list'
		], [
			[qw/authorize/],
			[
				'REQ MUL' => 'user'
			],
			'Authorizes [1] to add you to their buddie list'
		], [
			[qw/deauthorize/],
			[
				'REQ MUL' => 'user'
			],
			'Un-authorizes [1] to add you to their buddie list'
		], [
			[qw/block ban/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to block [1], ie stops them from talking to '.
			'you'
		], [
			[qw/unblock unban/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to unblock [1], ie lets them talk to you ' .
			'again'
		], [
			[qw/chat/],
			[
				'REQ MUL' => 'users'
			],
			'Attempts to start a chat with the [1]'
		], [
			[qw/clear/],
			[
				'OPT' => 'window number'
			],
			'Clears the screen of window [1] (or the current one '.
			'if none specified'
		], [
			[qw/invite/],
			[
				'REQ MUL' => 'users to invite'
			],
			'Invites [1] to the current chat session'
		], [
			[qw/change_fuser setname setfuser nick/],
			[
				'REQ' => 'new username'
			],
			'Changes your "friendly" username to [1]'
		], [
			[qw/file_get get/],
			[
				'REQ' => 'transfer number'
			],
			'Attempts to get the file that is waiting on [1]'
		], [
			[qw/file_noget noget cancel/],
			[
				'REQ' => 'transfer number'
			],
			'Attempts to cancel [1]'
		], [
			[qw/help/],
			[
				'OPT' => 'command'
			],
			'Displays this help, or help on [1]'
		], [
			[qw/ping/],
			[],
			'Pings yourself to check the connection is still alive'
		], [
			[qw/uptime/],
			[],
			'Displays how long the client has been running for'
		], [
			[qw/version/],
			[],
			'Gives you info about the client'
		], [
			[qw/refresh redraw/],
			[],
			'Refreshes the screen after some other program fucked '.
			'it up'
		], [
			[qw/list/],
			[],
			'Gives you a list of people who are not offline at the'.
			' moment, and their states'
		], [
			[qw/destroy/],
			[
				'OPT' => 'window number'
			],
			'Destroys the window number [1]'
		], [
			[qw/eval/],
			[
				'REQ MUL' => 'perl code'
			],
			'Evaluates the specified perl code and outputs any ' .
			'output'
		], [
			[qw/exec/],
			[
				'OPT' => '-o',
				'REQ MUL' => 'sh code'
			],
			'Evaluates the specified sh code and outputs any ' .
			'output to the current session if -o is specified, ' .
			'otherwise echos to the current screen.'
		]
	);

my %funcs;

# We now map all the possible functions into the keys hash, which conveniantly
# all point at the 'primary' function, ie, the one we want called.. :)
for(@funcs) {
	my $pri = $_->[0][0];

	for my $t (@{$_->[0]}) {
		$funcs{$t} = [$pri, $_]
	}
}

&init_screen;
&new_win(0);
&change_win(0);

# Restore the screen to original state if we get killed
$SIG{$_} = sub { &quit if $disconnect; $disconnect = 1 } for qw/KILL STOP TERM/,
								qw/INT/;

$SIG{WINCH} = sub { $needs_resize = 1 };

# Set a signal handler to stop the screen being messed up when something bad
# happens
$SIG{__WARN__} = sub { &out(undef, @_) };
$SIG{__DIE__} = sub { &out(undef, @_); &quit };

&handle_version;
&out(undef, 'Type /help to display help');
&out(undef, 'Attempting to connect...');

# Open the log file
open F, '>>log' or die "Can't open log: $!";
F->autoflush(1);

# Start the MSN module
my $msn = MSN->new(Debug => \*F) or &quit;

# Register this list of callbacks
my @callbacks = (
			qw/connect message new_mail open_chat disconnected/,
			qw/close_chat user_types join part change_state out/,
			qw/file_invite file_start error unknown_user/,
			qw/auth_success deauth_success add_success del_success/,
			qw/block_success unblock_success add del ping/,
			qw/state_success fuser_success/,
		);

eval '$msn->add_callback($_, \\&do_' . $_ . ')' for @callbacks;

$conf{idle} = 600;		# Default values
&read_config;
$conf{user} ||= &ask(0, 'Please enter your MSN username');
$conf{pass} ||= &ask(1, 'Please enter your MSN password');

my %opts = (
		User => $conf{user},
		Pass => $conf{pass},
	);

$opts{Host} = $conf{host} if exists $conf{host};
$opts{Port} = $conf{port} if exists $conf{port};
$opts{FUser} = $conf{fuser} if exists $conf{fuser};

# Try to connect!
if(!defined $msn->connect(%opts)) {
	&error("Can't connect");
	&quit
}

# We're connected
&update_status;

while(1) {
	$disconnect and &quit;
	$needs_resize and &resize_term;

	# Got anything to do with MSN?
	my $fh = $msn->do_stuff(undef, 1, [\*STDIN]);	# Check every second
	&quit unless defined $fh;

	if(defined(my $t = &in)) {
		if($t =~ m!^/!) {
			&func($t)
		} else {
			&handle_msg("$session $t")
		}
	} elsif(!$idle and $conf{idle} and $msn->get_state eq 'online') {	
		if(time - $last_type > $conf{idle}) {		# We idled!

			$msn->change_my_state('idle');
			&out(0, "Starting auto-idle after $conf{idle} seconds");
			$idle = 1;
		}
	}

	# Force-update the users ever second to see if they go idle
	&update_users if time > $last_user + 1;

	# Lag-timer
	if(defined $ping_last) {
		my $diff = time - $ping_last;

		# If we didn't get a ping after we sent the last, and we're 1
		# second more than when we sent the last, start incrementing
		# the lag counter till we get one back.. (We might have
		# disconnected)
		if($ping_got < $ping_last and $diff >= $PING_COUNT) {
			if($time_hires) {
				$last_ping = sprintf "%0.3f", $diff
			} else {
				$last_ping = int $diff
			}

			&update_status
		}

		&handle_ping if $diff >= $PING_SAMPLE
	}
}

sub handle_quit { &quit }
sub quit {
	&out(undef, "Disconnecting");
	&handle_uptime;
	$msn->disconnect if defined $msn;
	close F;
	$in_line->erase;
	$in_line->refresh;
	endwin;				# Close down the windowing system
	exit
}

sub func {
	my $cmd = shift;

	$cmd =~ s!^/!!;			# Remove leading / if from cmdline

	return &handle_msg($cmd) if $cmd =~ s/^\s+/$session /;

	my ($c, $s) = $cmd =~ /^(\w+)\s*(.*)$/;
	$c ||= '';

	my @matches = sort &complete($c, [keys %funcs]);
	if(@matches == 1) {			# 1 match
		eval '&handle_' . $funcs{$matches[0]}[0] . '($s)';
		&out(undef, $@) if $@;
		return
	} else {
		&list_commands($c)
	}
}

################################################################################
# Handle bits
################################################################################

sub handle_add		{ $msn->add_contact($_)		for split ' ', shift }
sub handle_del		{ $msn->del_contact($_)		for split ' ', shift }
sub handle_block	{ $msn->block_contact($_)	for split ' ', shift }
sub handle_unblock	{ $msn->unblock_contact($_)	for split ' ', shift }
sub handle_authorize	{ $msn->authorize_contact($_)	for split ' ', shift }
sub handle_deauthorize	{ $msn->deauthorize_contact($_)	for split ' ', shift }

sub handle_status	{ $msn->change_my_state(shift) }
sub handle_away		{ &handle_status('away') }
sub handle_online	{ &handle_status('online') }
sub handle_hide		{ &handle_status('hidden') }

sub handle_change_fuser	{ $msn->change_fuser(shift) }
sub handle_chat		{ $msn->start_ring(@_) }

sub handle_clear {
	my $win = shift;
	$win =~ /\D/ and $win = $session;
	return unless exists $sessions{$win};

	$sessions{$win}{Screen} = [];
	$sessions{$win}{Scroll} = 0;
	$sessions{$win}{Active} = 1;

	&update_win
}

sub handle_version {
	&out(undef, "jMSN version $MSN_VERSION by Mark Zealey",
							'(mark@zealos.org)');
}

sub handle_uptime {
	my ($sec, $min, $hours, $days);

	$days = time - $startup_time;

	$sec = $days % 60;
	$days = int($days / 60);
	$min = $days % 60;
	$days = int($days / 60);
	$hours = $days % 24;
	$days = int($days / 24);

	&out(undef, "Current uptime: ${days}d ${hours}h ${min}m ${sec}s");
}

sub handle_ping {
	# Don't send more than 1 ping at a time, so the lag counter works
	# properly when we're disconnected..
	return if defined $ping_last and $ping_got < $ping_last;

	$msn->ping;
	$ping_last = time
}

sub handle_part {
	my $str = shift;

	$str ||= $session;

	exists $sessions{$str} or return &out(undef, 'Parameters for bye are '.
								'invalid');

	$msn->part_sess($str) or return &error
}

sub handle_window { &change_win(shift) }

sub handle_destroy {
	my $num = shift;

	$num ||= $session;

	return &error("Unknown window number '$num'") if $num =~ /\D/;
	return &error("Can't destroy the main window") if $num == 0;

	$msn->part_sess($num);

	&change_win(0) if $session == $num;

	delete $sessions{$num}
}

sub handle_refresh { curscr->refresh }

sub handle_eval {
	my $code = shift;

	eval $code;

	&error($@) if $@
}

sub handle_exec {
	my $code = shift;

	$code .= ' 2>&1';

	if($code =~ s/^\s*-o\s+//) {		# Output to chat
		return &error("Can't output to main screen") if !$session;
		&handle_msg("$session " . `$code`)
	} else {
		&out(undef, `$code`)
	}
}

sub handle_file_get {
	my $cmd = shift;

	return if $cmd =~ /\D/;

	$msn->file_get($cmd, 1) or return;
	&out(undef, "Starting file transfer $cmd");
}

sub handle_file_noget {
	my $cmd = shift;

	return if $cmd =~ /\D/;

	$msn->file_get($cmd, 0) or return;
	&out(undef, "Canceling file transfer $cmd");
}

sub handle_invite {
	my $line = shift;

	my ($user, $sess) = $line =~ /^(\S+)\s*(\d+)?$/ or return
							&error('Invalid line');

	$sess ||= $session;
	
	$msn->invite_users($sess, $user);
}

sub handle_list {
	my %users = $msn->get_users;

	&out(undef, '');
	&out(undef, 'Users who are currently online:');

	for(keys %users) {
		my $s = $msn->get_human_state($_);
		next if $s eq 'offline';
		&out(undef, &get_fuser($users{$_}), "($_) => $s")
	}
}

sub handle_msg {
	my $cmd = shift;

	my ($to, $text) = $cmd =~ /^(\S+)\s*(.*)$/is or return;
	$last_msg = $to;
	if($msn->sess_exists($to)) {
		&out($to, &timestamp . '<' . $msn->plain_fuser($msn->{FUser}) .
								'>', $text);
		$msn->send_msg_swb($to, $text)
	}

	# Else maybe start new swb and msg them.
}

sub handle_help {
	my $cmd = shift || '';

	if($cmd =~ /(\S+)/) {			# Help on a specific command
		$cmd = $1;
		my @matches = &complete($cmd, [keys %funcs]);
		if(@matches == 0) {
			return &out(undef, "Invalid command '$cmd'")
		} elsif(@matches == 1) {		# Help for just this one
			my $help = $funcs{$matches[0]}[1];
			my $str = $help->[2];
			my @opts;

			# Walk through the arguments list and turn it into a
			# string...
			my ($arg, $num);
			for(@{$help->[1]}) {
				if($arg) {
					$num++;

					my ($start, $end);
					my @s = split ' ', $arg;
					if(grep {$_ eq 'OPT'} @s) {
						$start = '[';
						$end = ']'
					} else {
						$start = '<';
						$end = '>'
					}

					$str =~ s/\[$num]/${start}${_}${end}/g;

					if(grep {$_ eq 'MUL'} @s) {
						$end .= '...'
					}

					push @opts, $start . $_ . $end;

					undef $arg
				} else {
					$arg = $_
				}
			}

			&out(undef, "/$matches[0]", @opts);
			&out(undef, $str);
			return
		}
	}

	&list_commands($cmd);
}

################################################################################
# Callbacks
################################################################################

sub do_connect { &out(undef, 'Connected!'); $ping_last = 0 }

sub do_error { &error($_[1]) }

sub do_unknown_user {
	my ($msn, $user) = @_;
	&out($msn->get_sess_id, "Unknown user $user")
}

sub do_disconnected {
	&error('Disconnected!');
	&quit
}

sub do_new_mail {
	my ($msn, $msg) = @_;
	&out($msn->get_sess_id, "New mail from $msg->{From} <$msg->{FromAddr}>",
					"regarding $msg->{Subject}")
}

sub do_message {
	my ($msn, $msg) = @_;
	$sessions{$msn->get_sess_id}{Users}{$msg->{User}} = 0;
	&out($msn->get_sess_id, &timestamp . '<' . &get_fuser($msg) .
							"> $msg->{Body}")
}

sub do_ping {
	my ($msn, $time) = @_;

	$ping_got = time;

	if($time_hires) {
		$last_ping = sprintf "%0.3f", $time
	} else {
		$last_ping = int $time
	}

	&update_status
}

sub do_open_chat {
	my ($msn, $num) = @_;
	&new_win($num);
	&out($num, 'Chat session started');
}

sub do_close_chat {
	my ($msn, $num) = @_;
	&out($num, 'Chat session ended');
	&destroy_win($num);
}

sub do_join {
	my ($msn, $u) = @_;
	my $s = $msn->get_sess_id;
	$sessions{$s}{Users}{$u->{User}} = 0;
	$sessions{$s}{Active} = 1;				# Make active
	if($s == $session) {
		&update_win
	} else {
		&update_status
	}
}

sub do_part {
	my ($msn, $u) = @_;
	my $s = $msn->get_sess_id;
	delete $sessions{$s}{Users}{$u->{User}};
	$sessions{$s}{Active} = 1;				# Make active
	if($s == $session) {
		&update_win
	} else {
		&update_status
	}
}

sub do_user_types {
	my ($msn, $u) = @_;
	my $s = $msn->get_sess_id;
	$sessions{$s}{Users}{$u->{User}} = time;
	$sessions{$s}{Active} = 1;				# Make active
	if($s == $session) {
		&update_win
	} else {
		&update_status
	}
}

sub list_generic {
	my ($str, $msn, $u) = @_;
	&out(0, &get_fuser($u), "($u->{User}) $str")
}

sub list_generic_add	{ list_generic('has been ' . shift, @_) }

sub do_auth_success	{ list_generic_add('authorized', @_) }
sub do_deauth_success	{ list_generic_add('deauthorized', @_) }
sub do_add_success	{ list_generic_add('added', @_) }
sub do_del_success	{ list_generic_add('removed', @_) }
sub do_block_success	{ list_generic_add('blocked', @_) }
sub do_unblock_success	{ list_generic_add('unblocked', @_) }
sub do_del		{ list_generic('has removed you', @_) }

sub do_add {
	my ($msn, $u) = @_;
	list_generic('requests to be added', @_);
	&out(0, "Type /auth $u->{User} to allow them to add you")
}

sub do_state_success	{ &update_status }
sub do_fuser_success	{ &update_status }

sub do_change_state {
	my ($msn, $user) = @_;

	my $hst = &get_fuser($user);
        &out(0, "$hst ($user->{User}) has changed status to '" .
                                    $msn->get_human_state($user->{User}) . "'");
}

sub do_out {
	my ($msn, $reason) = @_;
	&out(undef, "Quitting: $reason");
	&quit
}

sub do_file_invite {
	my ($msn, $msg) = @_;

	&out($msn->get_sess_id, "$msg->{User} requests you to get file",
					"$msg->{File}, size $msg->{Size}");
	&out($msn->get_sess_id, "Type /get $msg->{Num} to get it");
}

sub do_file_start {
	my ($msn, $msg) = @_;

	# Debug I guess
	&out(0, "Transfer $msg->{Num} started, IP $msg->{Host}:$msg->{Port}");
}

################################################################################
# Windowing functions
################################################################################

sub init_screen {
	initscr;			# Curses initalization
	noecho;				# Don't echo the chars we recieve

	$Text::Wrap::columns = $COLS;

	# Init colors
	start_color;
	init_pair 1, COLOR_WHITE, COLOR_MAGENTA;
	init_pair 2, COLOR_WHITE, COLOR_RED;
	init_pair 3, COLOR_WHITE, COLOR_BLUE;    

	$MAGENTA = COLOR_PAIR(1);  
	$RED     = COLOR_PAIR(2);
	$BLUE    = COLOR_PAIR(3);

	# Main window where all the output goes to
	$win = newwin($LINES - 2, 0, 0, 0);

	$user_win = newwin(0, 0, 0, 0);

	# New window for reading input from
	$in_line = newwin(1, 0, $LINES - 1, 0);
	$in_line->nodelay(1);		# Non-blocking reading
	$in_line->keypad(1);		# Enable nice character mappings

	# New window for reading input from
	$stat_line = newwin(1, 0, $LINES - 2, 0);
	$stat_line->attron($BLUE);
}

sub new_win {
	my $num = shift;

	if(exists $sessions{$num}) {
		# Just add a blank line on if we have had a previous chat on
		# this window
		&out($num, '')
	} else {
		$sessions{$num} = {
					Screen => [],
					Users => {},
					Active => 1,
					Scroll => 0,
					Lines => $LINES,
				};
		&update_status
	}

	$sessions{$num}{Type} = 0
}

sub change_win {
	my $num = shift;

	return &out(undef, "Invalid window $num") unless exists $sessions{$num};
	$prompt = "[$num]" if !$ask;

	return if $num == $session;

	$sessions{$num}{Active} = 1;		# Needed.

	$session = $num;
	&update_win
}

sub destroy_win {
	my $win = shift;
	$sessions{$win}{Active} = 1;
	$sessions{$win}{Users} = {};
	&update_win
}

sub resize_term {
	$needs_resize = 0;
	my ($l, $c) = ($LINES, $COLS);

	# Euckie!
	for(`$RESIZE_PROG`) {
		/LINE.*=/i and ($LINES) = /(\d+)/;
		/COL.*=/i and ($COLS) = /(\d+)/;
	}

	endwin;
	initscr;

	$win->resize($LINES - 2, $COLS);

	$stat_line->mvwin($LINES - 2, 0);
	$stat_line->resize(1, $COLS);

	$in_line->mvwin($LINES - 1, 0);
	$in_line->resize(1, $COLS);

	&handle_refresh;

	&out(undef, "$l -> $LINES,  $c -> $COLS")
}

sub out {
	&out_block(shift, split "\n", wrap('', ' ' x 8, @_));
}

# window number, message. win == undef and puts to current window
sub out_block {
	my $num = shift;
	$num = $session if !defined $num;

	return print F "ACK! Bad window print!\n" unless exists $sessions{$num};

	my $w = $sessions{$num};

	push @{$w->{Screen}}, (@_?@_:'');

	# Fixup the 'drift'
	$w->{Scroll} += @_ if $w->{Scroll};
	$w->{Scroll} = $MAX_BUFFER_LEN - $w->{Lines} if
				$w->{Scroll} > $MAX_BUFFER_LEN - $w->{Lines};

	# Shrink the buffer if needed.
	my $t = @{$w->{Screen}} - $MAX_BUFFER_LEN;
	splice @{$w->{Screen}}, 0, $t, () if $t > 0;

	if($num == $session) {			# Current window, just update
		&update_win
	} elsif(!$w->{Active}) {
		# On another window, make active and inform luser
		$w->{Active} = 2;
		&update_status
	}

}

sub update_win {
	&update_users(1);

	my $w = $sessions{$session};

	my $start = @{$w->{Screen}} - $w->{Lines} - $w->{Scroll};
	$start = 0 if $start < 0;
	my $end = $#{$w->{Screen}} - $w->{Scroll};

	my $out = '';
	for(@{$w->{Screen}}[$start .. $end]) {
		if(length >= $COLS) {
			$out .= substr $_, 0, $COLS
		} else {
			$out .= "$_\n"
		}
	}

	$win->erase;
	$win->addstr($out);
	$win->refresh;

	if($w->{Active}) {
		$w->{Active} = 0;
		&update_status
	} else {					# The quick way
		$in_line->refresh
	}
}

sub update_users {
	my $t = shift;			# true if we are called from update_win
	my $w = $sessions{$session};
	my $n = keys %{$w->{Users}};

	return if !$n and !$t;		# Save work

	# Prevent being flooded if there are too many lusers here
	$n > $LINES / 2 and $n = $LINES / 2;

	$w->{Lines} = $LINES - $n - 2;

	$user_win->erase;
	$user_win->resize($n, $COLS) if $n;
	$win->mvwin($n, 0);
	$win->resize($w->{Lines}, $COLS);

	$last_user = time;

	for(keys %{$w->{Users}}) {
		last unless $n--;

		my $c = $w->{Users}{$_};

		my $str = "$_ - " . &get_fuser($msn->get_user($_));

		$str = substr $str, 0, $COLS;
		$str .= ' ' x ($COLS - length $str);

		my $col;

		my $d = $last_user - $c;
		if($c == 0) {
			$col = $BLUE
		} elsif($d >= 0 and $d <= 10) {	# Timeout of 10 seconds on type
			$col = $RED
		} elsif($d > 10) {
			$col = $MAGENTA
		}

		$user_win->attron($col);
		$user_win->addstr($str);
		$user_win->attroff($col)
	}

	$user_win->refresh if keys %{$w->{Users}};
	$in_line->refresh unless $t
}

sub update_status {
	if(defined $msn) {
		$status = "[$last_ping] " . $msn->plain_fuser($msn->{FUser}) . 
					" ($msn->{User}) - " . $msn->get_state;
		$status .= ' -- more --' if $sessions{$session}{Scroll};

		my @t;
		for(keys %sessions) {
			$_ == $session and $sessions{$_}{Active} = 0;
			push @t, $_ if $sessions{$_}{Active}
		}

		$status .= ' [' . join(',', @t) . ']' if @t;
	} else {					# Initializing...
		$status = 'Initializing';
	}

	$status = substr $status, 0, $COLS;
	$status .= ' ' x ($COLS - length $status);

	$stat_line->erase;
	$stat_line->addstr($status);
	$stat_line->refresh;
	&update_prompt
}

sub update_prompt {
	$in_line->erase;

	my $p = "$prompt ";
	my $curs = $char_ptr + length $p;

	$lines[$line_ptr] = '' if not defined $lines[$line_ptr];
	$p .= $lines[$line_ptr] unless $hide_input;

	# Make this cleverer..
	if($curs >= $COLS) {
		$p = substr $p, 1 - $COLS + $curs, $curs;
		$curs = $COLS
	}

	$in_line->addstr($p);
	$in_line->move(0, $curs);
	$in_line->refresh
}

################################################################################
# Input functions
################################################################################

# first param is wether to hide the input, the rest are the message
sub ask {
	my $old_hide_input = $hide_input;
	my $old_prompt = $prompt;
	my @old_lines = @lines;
	my $old_line_ptr = $line_ptr;

	$hide_input = shift;
	$ask = 1;
	$prompt = "@_";
	@lines = ('');
	$line_ptr = 0;

	&cbeep;
	&update_prompt;

	my $line;
	1 while not defined($line = &in);

	$hide_input = $old_hide_input;
	$prompt = $old_prompt;
	$ask = 0;
	@lines = @old_lines;
	$line_ptr = $old_line_ptr;
	&update_prompt;

	$line
}

# Return a full line, or undef if not avaliable (non-blocking)
sub in {
	while((my $t = $in_line->getch) ne ERR) {
		# Auto-idle stuff
		if($idle and defined $msn and $msn->get_state eq 'idle') {
			$idle = 0;
			$msn->change_my_state('online');
			&out(0, 'Unidleing after being idle for',
					int(time - $last_type), 'seconds')
		}

		$last_type = time;

		if($escape) {				# Alt-command
			$escape = 0;

			if($t =~ /^\d$/) {		# alt-num, change window
				&change_win(($t || 10) - 1)
			} elsif($t eq KEY_BACKSPACE or $t eq "\b") {
				my $i = substr $lines[$line_ptr], 0, $char_ptr;
				$i =~ s/([\w\d]*[^\w\d]*)$//;
				substr $lines[$line_ptr], 0, $char_ptr, $i;
				$char_ptr -= length $1
			}
		} elsif($t eq "\e") {			# Escape
			$escape = 1

		} elsif($t eq KEY_BACKSPACE or $t eq "\b") {
			substr $lines[$line_ptr], --$char_ptr, 1, ''
								if $char_ptr > 0
		} elsif($t eq KEY_DC) {
			substr $lines[$line_ptr], $char_ptr, 1, ''
		} elsif($t eq KEY_UP) {
			$line_ptr = 0 if ++$line_ptr >= @lines;
			$char_ptr = length $lines[$line_ptr]
		} elsif($t eq KEY_DOWN) {
			$line_ptr = 0 if --$line_ptr <= -@lines;
			$char_ptr = length $lines[$line_ptr]
		} elsif($t eq KEY_LEFT) {
			$char_ptr-- if $char_ptr > 0
		} elsif($t eq KEY_RIGHT) {
			$char_ptr++ if $char_ptr < length $lines[$line_ptr]
		} elsif($t eq KEY_HOME) {
			$char_ptr = 0
		} elsif($t eq KEY_END) {
			$char_ptr = length $lines[$line_ptr]
		} elsif($t eq KEY_PPAGE or $t eq KEY_NPAGE) {
			my $w = $sessions{$session};
			my $o = $w->{Scroll};
			$w->{Scroll} += int($w->{Lines} / 2) if $t eq KEY_PPAGE;
			$w->{Scroll} -= int($w->{Lines} / 2) if $t eq KEY_NPAGE;

			my $m = @{$w->{Screen}} - $w->{Lines};
			
			$w->{Scroll} = $m if $w->{Scroll} > $m;
			$w->{Scroll} = 0 if $w->{Scroll} < 0;

			next if $o == $w->{Scroll};	# Stop needless waste

			$w->{Active} = 1;

			&update_win
		} elsif($t eq KEY_IC) {
			$replace_on = !$replace_on
		} elsif($t eq "\014") {			# ^L
			&handle_refresh;

		} elsif(length($t) != 1) {		# Special char
			# Do nothing

		} elsif($t eq "\t") {{			# Tab-completion
			last if $ask;

			if($lines[$line_ptr] eq '') {
				# Just a <tab>, recall last msg command
				$lines[$line_ptr] = "/msg $last_msg ";
				$char_ptr = length $lines[$line_ptr];
				last
			}

			my $last = $lines[$line_ptr];
			($last) = substr($last, 0, $char_ptr) =~ /(\w*)$/;

			# XXX: tab-completion on different array-objects
			my @matches = sort &complete($last, [keys %funcs]);

			if(@matches == 1) {		# 1 match
				my $rem = substr($matches[0], length $last,
						length $matches[0]) . ' ';
				substr $lines[$line_ptr], $char_ptr, 0, $rem;
				$char_ptr += length $rem
			} elsif(@matches) {		# Lots of matches
				&list_commands($last)
			} else {			# No matches
				&out(undef, "No tab-match found for $last")
			}
		}} elsif($t eq "\n") {
			next if $lines[$line_ptr] eq '';
			$lines[0] = $t = $lines[$line_ptr];

			$char_ptr = 0;
			$line_ptr = 0;

			unshift @lines, '';

			pop @lines if @lines > $MAX_LINE_LEN;
			&update_prompt;

			# Reset the time we last typed.
			$sessions{$session}{Type} = 0;

			return $t
		} else {
			# Send a user-is-typing message to the swb if we havn't
			# in the last 4 seconds.
			my $ti = time;

			if($session and $ti >= $sessions{$session}{Type} + 4) {
				$sessions{$session}{Type} = $ti;
				$msn->send_typing_swb($session)
			}

			substr $lines[$line_ptr], $char_ptr++, 0, $t
		}

		&update_prompt
	}

	undef
}

################################################################################
# Useful functions
################################################################################

# Returns the timestamp in the format the luser wanted it in :-)
sub timestamp {
	return strftime($TIME_FORMAT, localtime) . ' ' if $TIME_FORMAT;
	''
}

sub read_config {
	# Try to read the config
	if(open G, $CONFIG) {
		my $config_file;
		while(<G>) {
			chomp;
			s/\s*#.*//g;
			next unless $_;
			$config_file .= "$_,"
		}
		close G;
		eval "\%conf = (\%conf, $config_file)"
	} else {
		&out(undef, "Warning: Can't open $CONFIG: $!")
	}
}

sub get_fuser {
	my $h = shift;
	my $nick = $msn->plain_fuser($msn->get_user($h->{User})->{FUser});
	$nick = substr $nick, 0, $MAX_FUSER_LEN if $MAX_FUSER_LEN;
	$nick
}

# Returns a command from a list
sub complete {
	my ($word, $words) = @_;

	return @$words unless defined $word;

	# First check complete matches
	for(@$words) {
		return ($_) if lc $_ eq lc $word
	}

	# Then check for partial matches and return
	grep /^$word/i, @$words;
}

sub list_commands {
	my $start = shift;

	$start ||= '';

	my %uniq;

	$uniq{$funcs{$_}[0]}++ for &complete($start, [keys %funcs]);

	return &out(undef, "No matches found for $start") unless keys %uniq;

	my @l;

	for my $t (map {$funcs{$_}[1]} sort keys %uniq) {
		my @commands = @{$t->[0]};
		if(@commands == 1) {			# 1 command
			push @l, $commands[0]
		} else {				# Multiple commands
			push @l, "{@commands}"
		}
	}

	&out_block(undef, 'Matches are:', map {"  $_"} @l);
}

sub cbeep { beep if $BEEP }

sub error {
	my $text = "@_";

	&cbeep;

	$text ||= $msn->{Error};
	&out(undef, "Error: $text")
}
