#!/usr/bin/perl -w
# Copyright Mark Zealey 2001. Released under the GPL.
use strict;
use MSN;
use IO::Handle;
use Term::ReadKey;
use constant DEFAULT_READ_MODE => 3;

open G, 'config' or die "Can't open config: $!";
my $config_file;
$config_file .= "$_," while <G>;
close G;
my %conf;
eval "\%conf = ($config_file)";

my $session = 0;			# Number of the current session
my $in = '';				# Line-buffer
my $last_msg = '';

my %funcs = (
		msg => \&handle_msg,
		window => \&handle_win,
		quit => \&quit,
		exit => \&quit,
		bye => \&handle_part,
		part => \&handle_part,
		add => \&handle_add,
		del => \&handle_del,
		chat => \&handle_chat,
	);

ReadMode DEFAULT_READ_MODE;		# Set to read-a-char-at-a-time-with-echo

open F, '>>log' or die "Can't open log: $!";
F->autoflush(1);

&out(undef, "Attempting to connect...");

$conf{user} ||= &ask(undef, 'Please enter your MSN username ');
$conf{pass} ||= &ask(undef, 'Please enter your MSN password ');

my $msn = MSN->new(
			Debug => \*F,
			User => $conf{user},
			Pass => $conf{pass},
	);

if(!defined $msn->connect) {
	&out(undef, "Error: $msn->{Error}");
	exit 1
}

my $disconnect = 0;

$SIG{INT} = sub { $disconnect = 1 };
&out(undef, "Connected!!");
$msn->add_callback('message', \&do_msg);
$msn->add_callback('new_mail', \&do_new_mail);
$msn->add_callback('add', \&do_add);
$msn->add_callback('new_user', \&do_new_user);
$msn->add_callback('open_chat', \&do_open_chat);
$msn->add_callback('close_chat', \&do_close_chat);
$msn->add_callback('join', \&do_join);
$msn->add_callback('part', \&do_part);
$msn->add_callback('change_state', \&do_change_state);

while(1) {
	$disconnect && &quit;

	my $t;
	if($t = &in) {
		if($t =~ m!^/!) {
			&func($t)
		} else {
			&handle_msg("$session $t")
		}
	}

	# Got anything to do with MSN?
	defined $msn->do_stuff(1, undef, \*STDIN) or $disconnect = 1;
}

sub quit {
	&out(undef, "Disconnecting");
	ReadMode 0;
	$msn->disconnect;
	close F;
	exit
}

sub func {
	my $cmd = shift;

	$cmd =~ s!^/!!;			# Remove leading / if from cmdline

	return &handle_msg($cmd) if $cmd =~ s/^\s+/$session /;

	my ($c, $s) = $cmd =~ /^(\w+)\s*(.*)$/;

	my @matches = sort &complete($c, [keys %funcs]);
	return &{$funcs{$matches[0]}}($2) if @matches == 1;

	if(@matches) {			# Multiple matches
		&out(undef, "Matches for $c:");
		&out(undef, "  $_") for @matches
	} else {			# No matches
		&out(undef, "Command $1 is unknown")
	}
}

sub handle_add {
	my $str = shift;

	$msn->add_contact(split ' ', $str) or return &out(undef,
							"Failed: $msn->{Error}")
}

sub handle_del {
	my $str = shift;

	my $u = $msn->get_user($str);
	my $s = '*** User ' . &get_fuser($u) . " ($u->{User}) has been removed";

	$msn->del_contact($str) or return &out(undef, "Failed: $msn->{Error}");
	&out(undef, $s)
}

sub handle_chat {
	my $user = shift;
	$msn->start_ring($user);
}

sub handle_part {
	my $str = shift;

	$str = $session unless $str;

	$str =~ /^\d+$/ or return &out(undef, 'Parameters for bye are invalid');

	$msn->part_sess($str) or return &out(undef, "Error: $msn->{Error}");
	&out($str, '*** You have left')
}

sub handle_win {
	my $cmd = shift;

	if($cmd =~ /^\d+$/ and $msn->sess_exists($cmd)) {
		$session = $cmd
	} else {
		&error("Invalid window number...")
	}
}

sub handle_msg {
	my $cmd = shift;

	my ($to, $text) = $cmd =~ /^(\S+)\s*(.*)$/i or return;
	$last_msg = $to;
	if($to =~ /^\d+$/ and $msn->sess_exists($to)) {
		&out($to, $text);
		$msn->send_msg_swb($to, $text)
	}

	# Else maybe start new swb and msg them.
}

sub do_online {
	my ($msn, $info) = @_;
	&out(undef, '*** ' . &get_fuser($info) . " ($info->{User}) has come " .
								'online');
}

sub do_offline {
	my ($msn, $info) = @_;
	&out(undef, '*** ' . &get_fuser($info) . " ($info->{User}) has gone " .
								'offline');
}

sub do_new_mail {
	my ($msn, $msg) = @_;
	&out($msn->get_sess_id, "New mail from $msg->{From} <$msg->{FromAddr}>".
					" regarding $msg->{Subject}")
}

sub do_msg {
	my ($msn, $msg) = @_;
	&out($msn->get_sess_id, '<' . &get_fuser($msg) . "> $msg->{Body}");
}

sub do_open_chat {
	my ($msn, $num) = @_;
	&out(undef, "*** Chat session $num started");
}

sub do_close_chat {
	my ($msn, $num) = @_;
	&out(undef, "*** Chat session $num ended");
}

sub do_join {
	my ($msn, $info) = @_;
	my $u = $msn->get_user($info);
	$session ||= $msn->get_sess_id;
	&out($msn->get_sess_id, '*** ' . &get_fuser($u) . " ($u->{User})".
						' has joined');
}

sub do_part {
	my ($msn, $info) = @_;
	my $u = $msn->get_user($info);
	&out($msn->get_sess_id, '*** ' . &get_fuser($u) . " ($u->{User})".
						' has left');
}

sub do_add {
	my ($msn, $user) = @_;
	my $u = $msn->get_user($user);
	&out(undef, '*** User ' . &get_fuser($u) . " ($u->{User}) has been " .
					'added');
}

sub do_new_user {
	my ($msn, $user) = @_;
	&out(undef, '*** User ' . &get_fuser($user) .
				" ($user->{User}) requests to be added");

	&ask(undef, "Add user (y/n)? ") =~ /^y/i and
					$msn->authorize_contact($user->{User});
}

sub do_change_state {
	my ($msn, $user) = @_;

	my $u = $msn->get_user($user);
	my $hst = &get_fuser($u);

	&out(undef, "*** User $hst ($u->{User}) has changed status to '" .
					$msn->get_human_state($user) . "'");
}

sub get_fuser {
	my $h = shift;
	$msn->plain_fuser($msn->get_user($h->{User})->{FUser})
}

# Returns a command from a list
sub complete {
	my ($word, $words) = @_;

	# First check complete matches
	for(@$words) {
		return ($_) if lc $_ eq lc $word
	}

	# Then check for partial matches and return
	grep /^$word/i, @$words;
}

# window number, message. win == undef and puts to current window
sub out {
	my $win = shift;
	$win = $session unless defined $win;
	print "[$win] @_\n";
}

sub ask {
	&out(@_);

	my $t;
	1 while !defined($t = &in);
	$t
}

sub tab_comp {
	if($in eq '') {		# Just a <tab>, recall last msg command
		$in = "/msg $last_msg ";
		print $in;
		return
	}

	my ($last) = $in =~ /(\w*)\s*$/;

	# XXX: tab-completion on different array-objects
	my @matches = sort &complete($last, [keys %funcs]);

	if(@matches == 1) {		# 1 match
		my $rem = substr $matches[0], length $last, 
					length $matches[0];
		print "$rem ";
		$in .= "$rem ";
	} elsif(@matches) {		# Lots of matches
		&out(undef, "Matches for $last:");
		&out(undef, "  $_") for @matches
	} else {			# No matches
		&out(undef, "No tab-match found for $last")
	}
}

sub in {
	my $t;

	while(defined($t = ReadKey(-1))) {
		if($t eq "\t") {		# Tab-completion
			&tab_comp;
		} elsif($t eq "\b" or ord $t == 127) {
			print "\b \b";
			chop $in
		} else {
			print $t;
			$in .= $t
		}
	}

	$in =~ s/(.*)\n//;
	$1
}

sub error { &out(undef, "Error: @_") }
